@@ -452,12 +452,12 @@ token_compare(const void *a, const void *b)
 }
 
 inline static sen_rc
-token_info_build(sen_index *i, const char *string, token_info **tis, uint32_t *n,
+token_info_build(sen_index *i, const char *string, size_t string_len, token_info **tis, uint32_t *n,
                  sen_sel_mode mode)
 {
   token_info *ti;
   sen_rc rc = sen_other_error;
-  sen_lex *lex = sen_lex_open(i->lexicon, string, 0);
+  sen_lex *lex = sen_lex_open(i->lexicon, string, string_len, 0);
   if (!lex) { return sen_memory_exhausted; }
   if (mode == sen_sel_unsplit) {
     if ((ti = token_info_open(i, (char *)lex->orig, 0, expand_both))) {
@@ -676,13 +676,13 @@ sen_index_create(const char *path, int key_size,
   strcpy(buffer, path);
   strcat(buffer, ".SEN");
   i->foreign_keyp = 0;
-  if ((i->keys = sen_sym_create(buffer, key_size, 0, sen_enc_none))) {
+  if ((i->keys = sen_sym_create(buffer, key_size, (flags & 0x70000), sen_enc_none))) {
     strcpy(buffer, path);
     strcat(buffer, ".SEN.l");
     if ((i->lexicon = sen_sym_create(buffer, 0, build_flags(flags), encoding))) {
       strcpy(buffer, path);
       strcat(buffer, ".SEN.i");
-      if ((i->inv = sen_inv_create(buffer, i->lexicon, 0, initial_n_segments))) {
+      if ((i->inv = sen_inv_create(buffer, i->lexicon, initial_n_segments))) {
         if ((flags & SEN_INDEX_WITH_STORE)) {
           strcpy(buffer, path);
           strcat(buffer, ".SEN.v");
@@ -774,7 +774,7 @@ sen_index_create_with_keys(const char *path, sen_sym *keys,
   if ((i->lexicon = sen_sym_create(buffer, 0, build_flags(flags), encoding))) {
     strcpy(buffer, path);
     strcat(buffer, ".SEN.i");
-    if ((i->inv = sen_inv_create(buffer, i->lexicon, 0, initial_n_segments))) {
+    if ((i->inv = sen_inv_create(buffer, i->lexicon, initial_n_segments))) {
       if ((flags & SEN_INDEX_WITH_STORE)) {
         strcpy(buffer, path);
         strcat(buffer, ".SEN.v");
@@ -937,9 +937,9 @@ sen_index_lock(sen_index *i, int timeout)
   uint32_t count;
   for (count = 0;; count++) {
     uint32_t lock;
-    SEN_ATOMIC_ADD_EX(&i->keys->header->lock, 1, lock);
+    SEN_ATOMIC_ADD_EX(i->keys->lock, 1, lock);
     if (lock) {
-      SEN_ATOMIC_ADD_EX(&i->keys->header->lock, -1, lock);
+      SEN_ATOMIC_ADD_EX(i->keys->lock, -1, lock);
       if (!timeout || (timeout > 0 && timeout == count)) { break; }
       usleep(1);
       continue;
@@ -953,14 +953,14 @@ sen_rc
 sen_index_unlock(sen_index *i)
 {
   uint32_t lock;
-  SEN_ATOMIC_ADD_EX(&i->keys->header->lock, -1, lock);
+  SEN_ATOMIC_ADD_EX(i->keys->lock, -1, lock);
   return sen_success;
 }
 
 sen_rc
 sen_index_clear_lock(sen_index *i)
 {
-  i->keys->header->lock = 0;
+  *i->keys->lock = 0;
   return sen_success;
 }
 
@@ -988,7 +988,7 @@ sen_index_path(sen_index *i, char *pathbuf, int bufsize)
 /* update */
 
 inline static sen_rc
-index_add(sen_index *i, const void *key, const char *value)
+index_add(sen_index *i, const void *key, const char *value, size_t value_len)
 {
   int hint;
   sen_set *h;
@@ -1001,7 +1001,7 @@ index_add(sen_index *i, const void *key, const char *value)
   sen_store_buf *sbuf = NULL;
   //  sen_log("add > (%x:%x)", i, key);
   if (!(rid = sen_sym_get(i->keys, key))) { return sen_invalid_argument; }
-  if (!(lex = sen_lex_open(i->lexicon, value, SEN_LEX_ADD|SEN_LEX_UPD))) {
+  if (!(lex = sen_lex_open(i->lexicon, value, value_len, SEN_LEX_ADD|SEN_LEX_UPD))) {
     return sen_memory_exhausted;
   }
   if (i->store) { sbuf = sen_store_buf_open(len); }
@@ -1009,7 +1009,7 @@ index_add(sen_index *i, const void *key, const char *value)
   if (!h) {
     SEN_LOG(sen_log_alert, "sen_set_open on index_add failed !");
     sen_lex_close(lex);
-    return sen_memory_exhausted; 
+    return sen_memory_exhausted;
   }
   while (!lex->status) {
     (tid = sen_lex_next(lex));
@@ -1057,7 +1057,7 @@ exit:
 }
 
 inline static sen_rc
-index_del(sen_index *i, const void *key, const char *value)
+index_del(sen_index *i, const void *key, const char *value, size_t value_len)
 {
   sen_set *h;
   sen_lex *lex;
@@ -1070,7 +1070,7 @@ index_del(sen_index *i, const void *key, const char *value)
     return sen_invalid_argument;
   }
   // sen_log("del > (%x:%d:%d) %d", key, rid, *((int *)key), strlen(value));
-  if (!(lex = sen_lex_open(i->lexicon, value, SEN_LEX_UPD))) {
+  if (!(lex = sen_lex_open(i->lexicon, value, value_len, SEN_LEX_UPD))) {
     return sen_memory_exhausted;
   }
   h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
@@ -1121,7 +1121,6 @@ sen_index_upd(sen_index *i, const void *key,
               const char *oldvalue, unsigned int oldvalue_len,
               const char *newvalue, unsigned int newvalue_len)
 {
-  // TODO: use oldvalue_len/newvalue_len
   // if (!strcmp(oldvalue, newvalue)) { return sen_success; }
   sen_rc rc;
   if (!i || !key) { SEN_LOG(sen_log_warning, "sen_index_upd: invalid argument"); return sen_invalid_argument; }
@@ -1131,13 +1130,13 @@ sen_index_upd(sen_index *i, const void *key,
   }
   // sen_log("index_upd %p %x > '%s'->'%s'", i, key, oldvalue, newvalue);
   if (oldvalue && *oldvalue) {
-    if ((rc = index_del(i, key, oldvalue))) {
+    if ((rc = index_del(i, key, oldvalue, oldvalue_len))) {
       SEN_LOG(sen_log_error, "index_del on sen_index_upd failed !");
       goto exit;
     }
   }
   if (newvalue && *newvalue) {
-    rc = index_add(i, key, newvalue);
+    rc = index_add(i, key, newvalue, newvalue_len);
   } else {
     rc = sen_sym_del(i->keys, key);
   }
@@ -1185,7 +1184,6 @@ sen_values_close(sen_values *v)
 sen_rc
 sen_values_add(sen_values *v, const char *str, unsigned int str_len, unsigned int weight)
 {
-  // TODO:use str_len
   sen_value *vp;
   if (!v || !str) { SEN_LOG(sen_log_warning, "sen_values_add: invalid argument"); return sen_invalid_argument; }
   if (!(v->n_values & (INITIAL_VALUE_SIZE - 1))) {
@@ -1196,6 +1194,7 @@ sen_values_add(sen_values *v, const char *str, unsigned int str_len, unsigned in
   }
   vp = &v->values[v->n_values];
   vp->str = str;
+  vp->str_len = str_len;
   vp->weight = weight;
   v->n_values++;
   return sen_success;
@@ -1234,7 +1233,7 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
       goto exit;
     }
     for (j = newvalues->n_values, v = newvalues->values; j; j--, v++) {
-      if ((lex = sen_lex_open(i->lexicon, v->str, SEN_LEX_ADD|SEN_LEX_UPD))) {
+      if ((lex = sen_lex_open(i->lexicon, v->str, v->str_len, SEN_LEX_ADD|SEN_LEX_UPD))) {
         while (!lex->status) {
           if ((tid = sen_lex_next(lex))) {
             sen_set_get(new, &tid, (void **) &u);
@@ -1279,7 +1278,7 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
       goto exit;
     }
     for (j = oldvalues->n_values, v = oldvalues->values; j; j--, v++) {
-      if ((lex = sen_lex_open(i->lexicon, v->str, SEN_LEX_UPD))) {
+      if ((lex = sen_lex_open(i->lexicon, v->str, v->str_len, SEN_LEX_UPD))) {
         while (!lex->status) {
           if ((tid = sen_lex_next(lex))) {
             sen_set_get(old, &tid, (void **) &u);
@@ -1350,7 +1349,10 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
     sen_set_cursor_close(c);
     sen_set_close(new);
   } else {
-    rc = sen_sym_del(i->keys, key);
+    if (!section) {
+      /* todo: delete key when all sections deleted */
+      rc = sen_sym_del(i->keys, key);
+    }
   }
 exit :
   sen_index_unlock(i);
@@ -2124,7 +2126,8 @@ get_weight(sen_records *r, sen_id rid, int sid,
 
 sen_rc
 sen_index_similar_search(sen_index *i, const char *string,
-                         sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
+                         unsigned int string_len, sen_records *r,
+                         sen_sel_operator op, sen_select_optarg *optarg)
 {
   int *w1, limit;
   sen_id tid, *tp;
@@ -2135,7 +2138,7 @@ sen_index_similar_search(sen_index *i, const char *string,
   if (!(h = sen_set_open(sizeof(sen_id), sizeof(int), 0))) {
     return sen_memory_exhausted;
   }
-  if (!(lex = sen_lex_open(i->lexicon, string, 0))) {
+  if (!(lex = sen_lex_open(i->lexicon, string, string_len, 0))) {
     sen_set_close(h);
     return sen_memory_exhausted;
   }
@@ -2237,11 +2240,13 @@ sen_index_similar_search(sen_index *i, const char *string,
 
 sen_rc
 sen_index_term_extract(sen_index *i, const char *string,
-                       sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
+                       unsigned int string_len, sen_records *r,
+                       sen_sel_operator op, sen_select_optarg *optarg)
 {
   posinfo pi;
   sen_id tid;
   const char *p;
+  sen_nstr *nstr;
   sen_inv_cursor *c;
   sen_inv_posting *pos;
   sen_rc rc = sen_success;
@@ -2249,7 +2254,10 @@ sen_index_term_extract(sen_index *i, const char *string,
   sen_wv_mode wvm = sen_wv_none;
   int skip, position, rep, policy;
   char keybuf[SEN_SYM_MAX_KEY_SIZE];
-  if (!i || !string || !r || !optarg) { return sen_invalid_argument; }
+  if (!i || !string || !string_len || !r || !optarg ||
+      !(nstr = sen_nstr_open(string, string_len, sym->encoding, 0)) ) {
+    return sen_invalid_argument;
+  }
   policy = optarg->max_interval;
   if (optarg->func) {
     wvm = sen_wv_dynamic;
@@ -2261,12 +2269,14 @@ sen_index_term_extract(sen_index *i, const char *string,
   }
   r->keys = i->keys;
   rep = (r->record_unit == sen_rec_position || r->subrec_unit == sen_rec_position);
-  for (p = string, position = 0; *p; p += skip, position += skip) {
+  for (p = nstr->norm, position = 0; *p; p += skip, position += skip) {
     if ((tid = sen_sym_common_prefix_search(sym, p))) {
       if (policy == TERM_EXTRACT_EACH_POST) {
         skip = sen_sym_key(sym, tid, keybuf, SEN_SYM_MAX_KEY_SIZE) - 1;
       } else {
-        skip = (int)sen_str_charlen(p, sym->encoding);
+        if (!(skip = (int)sen_str_charlen(p, sym->encoding))) {
+          break;
+        }
       }
       if (!(c = sen_inv_cursor_open(i->inv, tid, rep))) {
         SEN_LOG(sen_log_error, "cursor open failed (%d)", tid);
@@ -2295,7 +2305,9 @@ sen_index_term_extract(sen_index *i, const char *string,
       }
       sen_inv_cursor_close(c);
     } else {
-      skip = (int)sen_str_charlen(p, sym->encoding);
+      if (!(skip = (int)sen_str_charlen(p, sym->encoding))) {
+        break;
+      }
     }
   }
   sen_records_cursor_clear(r);
@@ -2306,6 +2318,7 @@ sen_index_term_extract(sen_index *i, const char *string,
     opt.compar_arg = (void *)(intptr_t)r->records->key_size;
     sen_records_sort(r, 10000, &opt);
   }
+  sen_nstr_close(nstr);
   return rc;
 }
 
@@ -2313,7 +2326,6 @@ sen_rc
 sen_index_select(sen_index *i, const char *string, unsigned int string_len,
                  sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
 {
-  // TODO: use string_len
   btr *bt = NULL;
   sen_rc rc = sen_success;
   int rep, orp, weight, max_interval = 0;
@@ -2331,18 +2343,18 @@ sen_index_select(sen_index *i, const char *string, unsigned int string_len,
     }
   }
   if (mode == sen_sel_similar) {
-    return sen_index_similar_search(i, string, r, op, optarg);
+    return sen_index_similar_search(i, string, string_len, r, op, optarg);
   }
   if (mode == sen_sel_term_extract) {
-    return sen_index_term_extract(i, string, r, op, optarg);
+    return sen_index_term_extract(i, string, string_len, r, op, optarg);
   }
   rep = (r->record_unit == sen_rec_position || r->subrec_unit == sen_rec_position);
   orp = (r->record_unit == sen_rec_position || op == sen_sel_or);
-  if (!(tis = SEN_MALLOC(sizeof(token_info *) * strlen(string) * 2))) {
+  if (!(tis = SEN_MALLOC(sizeof(token_info *) * string_len * 2))) {
     return sen_memory_exhausted;
   }
   r->keys = i->keys;
-  if (token_info_build(i, string, tis, &n, mode) || !n) { goto exit; }
+  if (token_info_build(i, string, string_len, tis, &n, mode) || !n) { goto exit; }
   if (mode == sen_sel_near) {
     if (!(bt = bt_open(n))) { rc = sen_memory_exhausted; goto exit; }
     max_interval = optarg->max_interval;
