@@ -27,7 +27,7 @@
 
 void sen_inv_seg_expire08(sen_inv *inv);
 sen_inv * sen_inv_create08(const char *path,  sen_sym *lexicon, 
-                           int flags, uint32_t initial_n_segments);
+                           uint32_t initial_n_segments);
 sen_inv * sen_inv_open08(const char *path, sen_sym *lexicon);
 sen_rc sen_inv_update08(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h,
                         int hint);
@@ -673,7 +673,6 @@ sen_inv_updspec_close(sen_inv_updspec *u)
 inline static uint8_t *
 encode_rec(sen_inv_updspec *u, unsigned int *size, int deletep)
 {
-  intptr_t s;
   uint8_t *br, *p;
   struct _sen_inv_pos *pp;
   uint32_t lpos, tf = deletep ? 0 : u->tf;
@@ -692,8 +691,8 @@ encode_rec(sen_inv_updspec *u, unsigned int *size, int deletep)
   for (lpos = 0, pp = u->pos; pp && tf--; lpos = pp->pos, pp = pp->next) {
     SEN_B_ENC(pp->pos - lpos, p);
   }
-  s = (p - br) + sizeof(buffer_rec);
-  *size = (unsigned int) ((s + 0x03) & ~0x03);
+  while (((intptr_t)p & 0x03)) { *p++ = 0; }
+  *size = (unsigned int) ((p - br) + sizeof(buffer_rec));
   return br;
 }
 
@@ -878,7 +877,7 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
   if ((scn = sb->header.chunk) != CHUNK_NOT_ASSIGNED) {
     sc = sen_io_win_map(inv->chunk, &sw, scn, 0, sb->header.chunk_size, sen_io_rdwr);
     if (!sc) {
-      SEN_LOG(sen_log_alert, "io_win_map(%d, %d) failed!!", scn, sb->header.chunk_size);
+      SEN_LOG(sen_log_alert, "io_win_map(%d, %d) failed!", scn, sb->header.chunk_size);
       buffer_close(inv, seg * SEN_INV_SEGMENT_SIZE);
       SEN_IO_SEG_UNREF(inv->seg, ds);
       SEN_FREE(tc);
@@ -1128,19 +1127,17 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
     db->header.nterms_void = nterms_void;
     inv->header->total_chunk_size += db->header.chunk_size >> 10;
   }
-  db->header.chunk = dcn;
+  db->header.chunk = db->header.chunk_size ? dcn : CHUNK_NOT_ASSIGNED;
   db->header.buffer_free = SEN_INV_SEGMENT_SIZE
     - sizeof(buffer_header) - sb->header.nterms * sizeof(buffer_term);
   db->header.nterms = sb->header.nterms;
 
   {
     uint32_t mc, ec;
-    mc = max_dest_chunk_size / SEN_INV_CHUNK_SIZE;
-    if (mc * SEN_INV_CHUNK_SIZE < max_dest_chunk_size) { mc++; }
-    ec = db->header.chunk_size / SEN_INV_CHUNK_SIZE;
-    if (ec * SEN_INV_CHUNK_SIZE < db->header.chunk_size) { ec++; }
+    mc = (max_dest_chunk_size + SEN_INV_CHUNK_SIZE - 1) / SEN_INV_CHUNK_SIZE;
+    ec = (db->header.chunk_size + SEN_INV_CHUNK_SIZE - 1) / SEN_INV_CHUNK_SIZE;
     while (ec < mc) {
-      inv->header->chunks[db->header.chunk + ec++] = 0;
+      inv->header->chunks[dcn + ec++] = 0;
     }
   }
   buffer_close(inv, seg * SEN_INV_SEGMENT_SIZE);
@@ -1159,8 +1156,7 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
 /* inv */
 
 sen_inv *
-sen_inv_create(const char *path,  sen_sym *lexicon, 
-               int flags, uint32_t initial_n_segments)
+sen_inv_create(const char *path, sen_sym *lexicon, uint32_t initial_n_segments)
 {
   int i, max_chunk;
   sen_io *seg, *chunk;
@@ -1168,7 +1164,7 @@ sen_inv_create(const char *path,  sen_sym *lexicon,
   char path2[PATH_MAX];
   struct sen_inv_header *header;
   if ((lexicon->flags & 0x70000)) {
-    return sen_inv_create08(path, lexicon, flags, initial_n_segments);
+    return sen_inv_create08(path, lexicon, initial_n_segments);
   }
   if (strlen(path) + 6 >= PATH_MAX) { return NULL; }
   strcpy(path2, path);
