@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -23,7 +23,7 @@
 /* ngram */
 
 inline static sen_lex *
-sen_ngram_open(sen_sym *sym, const char *str, uint8_t flags)
+sen_ngram_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
 {
   sen_lex *lex;
   if (!(lex = SEN_MALLOC(sizeof(sen_lex)))) { return NULL; }
@@ -38,19 +38,12 @@ sen_ngram_open(sen_sym *sym, const char *str, uint8_t flags)
   lex->flags = flags;
   lex->status = sen_lex_doing;
   lex->encoding = sym->encoding;
-  if (sym->flags & SEN_INDEX_NORMALIZE) {
-    if ((lex->nstr = sen_nstr_open(str, lex->encoding,
-				   SEN_STR_REMOVEBLANK|SEN_STR_WITH_CTYPES))) {
-      str = lex->nstr->norm;
-    }
-  } else {
-    lex->nstr = NULL;
-  }
-  lex->orig = (unsigned char *)str;
-  lex->next = (unsigned char *)str;
-  lex->uni_alpha = (lex->nstr && !(lex->sym->flags & SEN_INDEX_SPLIT_ALPHA));
-  lex->uni_digit = (lex->nstr && !(lex->sym->flags & SEN_INDEX_SPLIT_DIGIT));
-  lex->uni_symbol = (lex->nstr && !(lex->sym->flags & SEN_INDEX_SPLIT_SYMBOL));
+  lex->nstr = nstr;
+  lex->orig = (unsigned char *)nstr->norm;
+  lex->next = (unsigned char *)nstr->norm;
+  lex->uni_alpha = (nstr->ctypes && !(lex->sym->flags & SEN_INDEX_SPLIT_ALPHA));
+  lex->uni_digit = (nstr->ctypes && !(lex->sym->flags & SEN_INDEX_SPLIT_DIGIT));
+  lex->uni_symbol = (nstr->ctypes && !(lex->sym->flags & SEN_INDEX_SPLIT_SYMBOL));
   return lex;
 }
 
@@ -65,100 +58,143 @@ sen_ngram_next(sen_lex *lex)
   const unsigned char *p, *q, *r;
   if (lex->status) { return SEN_SYM_NIL; }
   for (p = lex->next, pos = lex->pos + lex->skip; *p; p = r, pos++) {
-    if (lex->nstr) { cp = lex->nstr->ctypes + pos; }
+    if (lex->nstr->ctypes) { cp = lex->nstr->ctypes + pos; }
     if (lex->uni_alpha && SEN_NSTR_CTYPE(*cp) == sen_str_alpha) {
       for (len = 1, r = p;;len++) {
-	r += sen_str_charlen((char *)r, lex->encoding);
-	if (SEN_NSTR_ISBLANK(*cp)) { break; }
-	if (!*r || SEN_NSTR_CTYPE(*++cp) != sen_str_alpha) { break; }
+        size_t cl;
+        if (!(cl = sen_str_charlen((char *)r, lex->encoding))) { break; }
+        r += cl;
+        if (SEN_NSTR_ISBLANK(*cp)) { break; }
+        if (SEN_NSTR_CTYPE(*++cp) != sen_str_alpha) { break; }
       }
-      if (!(buf = SEN_MALLOC(r - p + 1))) { return SEN_SYM_NIL; }
-      memcpy(buf, p, r - p);
-      buf[r - p] = '\0';
-      tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-      SEN_FREE(buf);
-      if (tid == SEN_SYM_NIL) {
-	lex->status = sen_lex_not_found;
-	return SEN_SYM_NIL;
+      {
+        size_t blen = r - p;
+        if (!blen) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
+        memcpy(buf, p, blen);
+        buf[blen] = '\0';
+        // sen_log("alpha(%s)", buf);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
+        SEN_FREE(buf);
+        if (tid == SEN_SYM_NIL) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        lex->skip = len;
       }
-      lex->skip = len;
     } else if (lex->uni_digit && SEN_NSTR_CTYPE(*cp) == sen_str_digit) {
       for (len = 1, r = p;;len++) {
-	r += sen_str_charlen((char *)r, lex->encoding);
-	if (SEN_NSTR_ISBLANK(*cp)) { break; }
-	if (!*r || SEN_NSTR_CTYPE(*++cp) != sen_str_digit) { break; }
+        size_t cl;
+        if (!(cl = sen_str_charlen((char *)r, lex->encoding))) { break; }
+        r += cl;
+        if (SEN_NSTR_ISBLANK(*cp)) { break; }
+        if (SEN_NSTR_CTYPE(*++cp) != sen_str_digit) { break; }
       }
-      if (!(buf = SEN_MALLOC(r - p + 1))) { return SEN_SYM_NIL; }
-      memcpy(buf, p, r - p);
-      buf[r - p] = '\0';
-      tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-      SEN_FREE(buf);
-      if (tid == SEN_SYM_NIL) {
-	lex->status = sen_lex_not_found;
-	return SEN_SYM_NIL;
+      {
+        size_t blen = r - p;
+        if (!blen) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
+        memcpy(buf, p, blen);
+        buf[blen] = '\0';
+        // sen_log("digit(%s) len=%d", buf, len);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
+        SEN_FREE(buf);
+        if (tid == SEN_SYM_NIL) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        lex->skip = len;
       }
-      lex->skip = len;
     } else if (lex->uni_symbol && SEN_NSTR_CTYPE(*cp) == sen_str_symbol) {
       for (len = 1, r = p;;len++) {
-	r += sen_str_charlen((char *)r, lex->encoding);
-	if (SEN_NSTR_ISBLANK(*cp)) { break; }
-	if (!*r || SEN_NSTR_CTYPE(*++cp) != sen_str_symbol) { break; }
+        size_t cl;
+        if (!(cl = sen_str_charlen((char *)r, lex->encoding))) { break; }
+        r += cl;
+        if (SEN_NSTR_ISBLANK(*cp)) { break; }
+        if (SEN_NSTR_CTYPE(*++cp) != sen_str_symbol) { break; }
       }
-      if (!(buf = SEN_MALLOC(r - p + 1))) { return SEN_SYM_NIL; }
-      memcpy(buf, p, r - p);
-      buf[r - p] = '\0';
-      tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-      SEN_FREE(buf);
-      if (tid == SEN_SYM_NIL) {
-	lex->status = sen_lex_not_found;
-	return SEN_SYM_NIL;
+      {
+        size_t blen = r - p;
+        if (!blen) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
+        memcpy(buf, p, blen);
+        buf[blen] = '\0';
+        // sen_log("symbol(%s)", buf);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
+        SEN_FREE(buf);
+        if (tid == SEN_SYM_NIL) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        lex->skip = len;
       }
-      lex->skip = len;
     } else {
+      size_t cl;
 #ifdef PRE_DEFINED_UNSPLIT_WORDS
       {
-	const unsigned char *key = NULL;
-	if ((tid = sen_sym_common_prefix_search(sym, p))) {
-	  key = _sen_sym_key(sym, tid);
-	  len = sen_str_len(key, lex->encoding, NULL);
-	}
-	r = p + sen_str_charlen(p, lex->encoding);
-	if (tid && (len > 1 || !*r)) {
-	  if (*r && pos + len - 1 <= lex->tail) { continue; }
-	  p += strlen(key);
-	  if (!*p && !(lex->flags & SEN_LEX_UPD)) { lex->status = sen_lex_done; }
-	}
+        const unsigned char *key = NULL;
+        if ((tid = sen_sym_common_prefix_search(sym, p))) {
+          key = _sen_sym_key(sym, tid);
+          len = sen_str_len(key, lex->encoding, NULL);
+        }
+        r = p + sen_str_charlen(p, lex->encoding);
+        if (tid && (len > 1 || r == p)) {
+          if (r != p && pos + len - 1 <= lex->tail) { continue; }
+          p += strlen(key);
+          if (!*p && !(lex->flags & SEN_LEX_UPD)) { lex->status = sen_lex_done; }
+        }
       }
 #endif /* PRE_DEFINED_UNSPLIT_WORDS */
-      r = p + sen_str_charlen((char *)p, lex->encoding);
+      if (!(cl = sen_str_charlen((char *)p, lex->encoding))) {
+        lex->status = sen_lex_not_found;
+        return SEN_SYM_NIL;
+      }
+      r = p + cl;
       {
-	int blankp = 0;
-	for (len = 1, q = r; len < SEN_LEX_NGRAM_UNIT_SIZE; len++) {
-	  if (cp) {
-	    if (SEN_NSTR_ISBLANK(*cp)) { blankp++; break; }
-	    cp++;
-	  }
-	  if (!*q ||
-	      (lex->uni_alpha && SEN_NSTR_CTYPE(*cp) == sen_str_alpha) ||
-	      (lex->uni_digit && SEN_NSTR_CTYPE(*cp) == sen_str_digit) ||
-	      (lex->uni_symbol && SEN_NSTR_CTYPE(*cp) == sen_str_symbol)) {
-	    break;
-	  }
-	  q += sen_str_charlen((char *)q, lex->encoding);
-	}
-	if (blankp && !(lex->flags & SEN_LEX_UPD)) { continue; }
+        int blankp = 0;
+        for (len = 1, q = r; len < SEN_LEX_NGRAM_UNIT_SIZE; len++) {
+          if (cp) {
+            if (SEN_NSTR_ISBLANK(*cp)) { blankp++; break; }
+            cp++;
+          }
+          if (!(cl = sen_str_charlen((char *)q, lex->encoding)) ||
+              (lex->uni_alpha && SEN_NSTR_CTYPE(*cp) == sen_str_alpha) ||
+              (lex->uni_digit && SEN_NSTR_CTYPE(*cp) == sen_str_digit) ||
+              (lex->uni_symbol && SEN_NSTR_CTYPE(*cp) == sen_str_symbol)) {
+            break;
+          }
+          q += cl;
+        }
+        if (blankp && !(lex->flags & SEN_LEX_UPD)) { continue; }
       }
-      if (!*q && !(lex->flags & SEN_LEX_UPD)) { lex->status = sen_lex_done; }
-      if (!(buf = SEN_MALLOC(q - p + 1))) { return SEN_SYM_NIL; }
-      memcpy(buf, p, q - p);
-      buf[q - p] = '\0';
-      tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-      SEN_FREE(buf);
-      if (tid == SEN_SYM_NIL) {
-	lex->status = sen_lex_not_found;
-	return SEN_SYM_NIL;
+      if ((!cl || !*q) && !(lex->flags & SEN_LEX_UPD)) { lex->status = sen_lex_done; }
+      {
+        size_t blen = q - p;
+        if (!blen) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
+        memcpy(buf, p, blen);
+        buf[blen] = '\0';
+        // sen_log("other(%s), pos=%d, %d", buf, pos, cp - lex->nstr->ctypes);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
+        SEN_FREE(buf);
+        if (tid == SEN_SYM_NIL) {
+          lex->status = sen_lex_not_found;
+          return SEN_SYM_NIL;
+        }
+        lex->skip = 1;
       }
-      lex->skip = 1;
     }
     lex->pos = pos;
     lex->len = len;
@@ -181,7 +217,7 @@ static mecab_t *sole_mecab;
 sen_mutex sole_mecab_lock;
 
 inline static sen_lex *
-sen_mecab_open(sen_sym *sym, const char *str, uint8_t flags)
+sen_mecab_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
 {
   unsigned int bufsize, maxtrial = 10, len;
   char *buf, *s, *p;
@@ -198,14 +234,8 @@ sen_mecab_open(sen_sym *sym, const char *str, uint8_t flags)
   lex->flags = flags;
   lex->status = sen_lex_doing;
   lex->encoding = sym->encoding;
-  if (sym->flags & SEN_INDEX_NORMALIZE) {
-    if ((lex->nstr = sen_nstr_open(str, lex->encoding, 0))) {
-      str = lex->nstr->norm;
-    }
-  } else {
-    lex->nstr = NULL;
-  }
-  len = (unsigned int)strlen(str);
+  lex->nstr = nstr;
+  len = nstr->norm_blen;
   for (bufsize = len * 2 + 1; maxtrial; bufsize *= 2, maxtrial--) {
     if(!(buf = SEN_MALLOC(bufsize + 1))) {
       SEN_LOG(sen_log_alert, "buffer allocation on sen_mecab_open failed !");
@@ -213,7 +243,7 @@ sen_mecab_open(sen_sym *sym, const char *str, uint8_t flags)
       return NULL;
     }
     MUTEX_LOCK(sole_mecab_lock);
-    s = mecab_sparse_tostr3(lex->mecab, (char *)str, len, buf, bufsize);
+    s = mecab_sparse_tostr3(lex->mecab, (char *)nstr->norm, len, buf, bufsize);
     MUTEX_UNLOCK(sole_mecab_lock);
     if (s) { break; }
     SEN_FREE(buf);
@@ -225,7 +255,7 @@ sen_mecab_open(sen_sym *sym, const char *str, uint8_t flags)
   }
   for (p = buf + strlen(buf) - 1; *p == '\n' || isspace(*(unsigned char *)p); p--) { *p = '\0'; }
   //sen_log("sparsed='%s'", s);
-  lex->orig = (unsigned char *)str;
+  lex->orig = (unsigned char *)nstr->norm;
   lex->buf = (unsigned char *)buf;
   lex->next = (unsigned char *)buf;
   return lex;
@@ -241,10 +271,13 @@ sen_mecab_next(sen_lex *lex)
   int32_t len, offset = lex->offset + lex->len;
   const unsigned char *p;
   if (lex->status) { return SEN_SYM_NIL; }
-  for (p = lex->next, len = 0; *p;) {
-    p += sen_str_charlen((char *)p, lex->encoding);
+  for (p = lex->next, len = 0;;) {
+    size_t cl;
+    if (!(cl = sen_str_charlen((char *)p, lex->encoding)) || isspace(*p)) {
+      break;
+    }
+    p += cl;
     len++;
-    if (isspace(*p)) { break; }
   }
   if (!len) {
     lex->status = sen_lex_done;
@@ -275,9 +308,10 @@ sen_mecab_next(sen_lex *lex)
 /* delimited */
 
 inline static sen_lex *
-sen_delimited_open(sen_sym *sym, const char *str, uint8_t flags)
+sen_delimited_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
 {
   sen_lex *lex;
+  const char *p;
   if (!(lex = SEN_MALLOC(sizeof(sen_lex)))) { return NULL; }
   lex->sym = sym;
 #ifndef NO_MECAB
@@ -290,19 +324,14 @@ sen_delimited_open(sen_sym *sym, const char *str, uint8_t flags)
   lex->flags = flags;
   lex->status = sen_lex_doing;
   lex->encoding = sym->encoding;
-  if (sym->flags & SEN_INDEX_NORMALIZE) {
-    if ((lex->nstr = sen_nstr_open(str, lex->encoding, 0))) {
-      str = lex->nstr->norm;
-    }
-  } else {
-    lex->nstr = NULL;
-  }
-  lex->orig = (unsigned char *)str;
-  while (isspace(*str)) { str++; }
-  lex->next = (unsigned char *)str;
+  lex->nstr = nstr;
+  p = nstr->norm;
+  lex->orig = (unsigned char *)p;
+  while (isspace(*p)) { p++; }
+  lex->next = (unsigned char *)p;
   lex->offset = 0;
   lex->len = 0;
-  if (!*str) { lex->status = sen_lex_done; }
+  if (!*p) { lex->status = sen_lex_done; }
   return lex;
 }
 
@@ -316,10 +345,13 @@ sen_delimited_next(sen_lex *lex)
   int32_t len, offset = lex->offset + lex->len;
   const unsigned char *p;
   if (lex->status) { return SEN_SYM_NIL; }
-  for (p = lex->next, len = 0; *p;) {
-    p += sen_str_charlen((char *)p, lex->encoding);
+  for (p = lex->next, len = 0;;) {
+    size_t cl;
+    if (!(cl = sen_str_charlen((char *)p, lex->encoding)) || isspace(*p)) {
+      break;
+    }
+    p += cl;
     len++;
-    if (isspace(*p)) { break; }
   }
   if (!len) {
     lex->status = sen_lex_done;
@@ -370,27 +402,47 @@ sen_lex_fin(void)
 }
 
 sen_lex *
-sen_lex_open(sen_sym *sym, const char *str, uint8_t flags)
+sen_lex_open(sen_sym *sym, const char *str, size_t str_len, uint8_t flags)
 {
-  switch ((sym->flags & SEN_INDEX_TOKENIZER_MASK)) {
+  sen_nstr *nstr;
+  int nflag, type;
+  if (!sym) {
+    SEN_LOG(sen_log_warning, "sym is null at sen_lex_open");
+    return NULL;
+  }
+  type = sym->flags & SEN_INDEX_TOKENIZER_MASK;
+  nflag = (type == SEN_INDEX_NGRAM ? SEN_STR_REMOVEBLANK|SEN_STR_WITH_CTYPES : 0);
+  if (sym->flags & SEN_INDEX_NORMALIZE) {
+    if (!(nstr = sen_nstr_open(str, str_len, sym->encoding, nflag))) {
+      SEN_LOG(sen_log_alert, "sen_nstr_open failed at sen_lex_open");
+      return NULL;
+    }
+  } else {
+    if (!(nstr = sen_fakenstr_open(str, str_len, sym->encoding, nflag))) {
+      SEN_LOG(sen_log_alert, "sen_fakenstr_open failed at sen_lex_open");
+      return NULL;
+    }
+  }
+  switch (type) {
   case SEN_INDEX_MORPH_ANALYSE :
 #ifdef NO_MECAB
     return NULL;
 #else /* NO_MECAB */
-    return sen_mecab_open(sym, str, flags);
+    return sen_mecab_open(sym, nstr, flags);
 #endif /* NO_MECAB */
   case SEN_INDEX_NGRAM :
-    return sen_ngram_open(sym, str, flags);
+    return sen_ngram_open(sym, nstr, flags);
   case SEN_INDEX_DELIMITED :
-    return sen_delimited_open(sym, str, flags);
+    return sen_delimited_open(sym, nstr, flags);
   default :
     return NULL;
-  }  
+  }
 }
 
 sen_rc
 sen_lex_next(sen_lex *lex)
 {
+  /* if (!lex) { return sen_invalid_argument; } */
   switch ((lex->sym->flags & SEN_INDEX_TOKENIZER_MASK)) {
   case SEN_INDEX_MORPH_ANALYSE :
 #ifdef NO_MECAB
@@ -404,7 +456,7 @@ sen_lex_next(sen_lex *lex)
     return sen_delimited_next(lex);
   default :
     return sen_invalid_argument;
-  }  
+  }
 }
 
 sen_rc
