@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -64,6 +64,142 @@ suffix_init(void)
   }
 }
 
+inline size_t
+sen_str_charlen_utf8(const unsigned char *str, const unsigned char *end)
+{
+  /* MEMO: This function allows non-null-terminated string as str. */
+  /*       But requires the end of string. */
+  const unsigned char *p = str;
+  if (!*p || p >= end) { return 0; }
+  if (*p & 0x80) {
+    int b, w;
+    size_t size;
+    for (b = 0x40, w = 0; b && (*p & b); b >>= 1, w++);
+    if (!w) {
+      SEN_LOG(sen_log_warning, "invalid utf8 string(1) on sen_str_charlen_utf8");
+      return 0;
+    }
+    for (size = 1; w--; size++) {
+      if (++p >= end || !*p || (*p & 0xc0) != 0x80) {
+        SEN_LOG(sen_log_warning, "invalid utf8 string(2) on sen_str_charlen_utf8");
+        return 0;
+      }
+    }
+    return size;
+  } else {
+    return 1;
+  }
+  return 0;
+}
+
+size_t
+sen_str_charlen(const char *str, sen_encoding encoding)
+{
+  /* MEMO: This function requires null-terminated string as str.*/
+  unsigned char *p = (unsigned char *) str;
+  if (!*p) { return 0; }
+  switch (encoding) {
+  case sen_enc_euc_jp :
+    if (*p & 0x80) {
+      if (*(p + 1)) {
+        return 2;
+      } else {
+        /* This is invalid character */
+        SEN_LOG(sen_log_warning, "invalid euc-jp string end on sen_str_charlen");
+        return 0;
+      }
+    }
+    return 1;
+    break;
+  case sen_enc_utf8 :
+    if (*p & 0x80) {
+      int b, w;
+      size_t size;
+      for (b = 0x40, w = 0; b && (*p & b); b >>= 1, w++);
+      if (!w) {
+        SEN_LOG(sen_log_warning, "invalid utf8 string(1) on sen_str_charlen");
+        return 0;
+      }
+      for (size = 1; w--; size++) {
+        if (!*++p || (*p & 0xc0) != 0x80) {
+          SEN_LOG(sen_log_warning, "invalid utf8 string(2) on sen_str_charlen");
+          return 0;
+        }
+      }
+      return size;
+    } else {
+      return 1;
+    }
+    break;
+  case sen_enc_sjis :
+    if (*p & 0x80) {
+      if (0xa1 <= *p && *p <= 0xdf) {
+        /* hankaku-kana */
+        return 1;
+      } else if (!(*(p + 1))) {
+        /* This is invalid character */
+        SEN_LOG(sen_log_warning, "invalid sjis string end on sen_str_charlen");
+        return 0;
+      } else {
+        return 2;
+      }
+    } else {
+      return 1;
+    }
+    break;
+  default :
+    return 1;
+    break;
+  }
+  return 0;
+}
+
+size_t
+sen_str_charlen_nonnull(const char *str, const char *end, sen_encoding encoding)
+{
+  /* MEMO: This function allows non-null-terminated string as str. */
+  /*       But requires the end of string. */
+  unsigned char *p = (unsigned char *) str;
+  if (p >= (unsigned char *)end) { return 0; }
+  switch (encoding) {
+  case sen_enc_euc_jp :
+    if (*p & 0x80) {
+      if ((p + 1) < (unsigned char *)end) {
+        return 2;
+      } else {
+        /* This is invalid character */
+        SEN_LOG(sen_log_warning, "invalid euc-jp string end on sen_str_charlen_nonnull");
+        return 0;
+      }
+    }
+    return 1;
+    break;
+  case sen_enc_utf8 :
+    return sen_str_charlen_utf8(p, (unsigned char *)end);
+    break;
+  case sen_enc_sjis :
+    if (*p & 0x80) {
+      if (0xa1 <= *p && *p <= 0xdf) {
+        /* hankaku-kana */
+        return 1;
+      } else if (++p >= (unsigned char *)end) {
+        /* This is invalid character */
+        SEN_LOG(sen_log_warning, "invalid sjis string end on sen_str_charlen_nonnull");
+        return 0;
+      } else {
+        return 2;
+      }
+    } else {
+      return 1;
+    }
+    break;
+  default :
+    return 1;
+    break;
+  }
+  return 0;
+}
+
 sen_rc
 sen_str_fin(void)
 {
@@ -143,7 +279,7 @@ sen_logger_pass(sen_log_level level)
 #define LBUFSIZE 0x400
 
 void
-sen_logger_put(sen_log_level level, 
+sen_logger_put(sen_log_level level,
                const char *file, int line, const char *func, char *fmt, ...)
 {
   if (level <= sen_logger->max_level) {
@@ -289,10 +425,10 @@ normalize_euc(sen_nstr *nstr)
     0xd1, 0, 0, 0xd4, 0, 0, 0xd7, 0, 0, 0xda, 0, 0, 0xdd
   };
   int_least8_t *ch;
-  const unsigned char *s, *s_;
+  const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_, b;
   uint_least8_t *cp, *ctypes, ctype;
-  size_t size = strlen(nstr->orig), length = 0;
+  size_t size = nstr->orig_blen, length = 0;
   int removeblankp = nstr->flags & SEN_STR_REMOVEBLANK;
   if (!(nstr->norm = SEN_MALLOC(size * 2 + 1))) {
     return sen_memory_exhausted;
@@ -313,9 +449,10 @@ normalize_euc(sen_nstr *nstr)
     }
   }
   cp = ctypes = nstr->ctypes;
-  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; *s; s++) {
+  e = (unsigned char *)nstr->orig + size;
+  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; s < e; s++) {
     if ((*s & 0x80)) {
-      if ((*(s + 1) & 0x80)) {
+      if (((s + 1) < e) && (*(s + 1) & 0x80)) {
         unsigned char c1 = *s++, c2 = *s, c3 = 0;
         switch (c1 >> 4) {
         case 0x08 :
@@ -497,6 +634,7 @@ normalize_euc(sen_nstr *nstr)
   if (cp) { *cp = sen_str_null; }
   *d = '\0';
   nstr->length = length;
+  nstr->norm_blen = (size_t)(d - (unsigned char *)nstr->norm);
   return sen_success;
 }
 
@@ -546,10 +684,10 @@ inline static sen_rc
 normalize_utf8(sen_nstr *nstr)
 {
   int_least8_t *ch;
-  const unsigned char *s, *s_, *p, *p2, *pe;
+  const unsigned char *s, *s_, *p, *p2, *pe, *e;
   unsigned char *d, *d0, *d_;
   uint_least8_t *cp, *ctypes;
-  size_t length = 0, ls, ls_ = 0, lp, size = strlen(nstr->orig);
+  size_t length = 0, ls, ls_ = 0, lp, size = nstr->orig_blen;
   int removeblankp = nstr->flags & SEN_STR_REMOVEBLANK;
   if (!(nstr->norm = SEN_MALLOC(size * 5 + 1))) { /* todo: realloc unless enough */
     return sen_memory_exhausted;
@@ -570,8 +708,11 @@ normalize_utf8(sen_nstr *nstr)
     }
   }
   cp = ctypes = nstr->ctypes;
-  for (s = s_ = (unsigned char *) nstr->orig, d = d0, d_ = NULL; *s; s += ls) {
-    ls = sen_str_charlen((char *)s, sen_enc_utf8);
+  e = (unsigned char *)nstr->orig + size;
+  for (s = s_ = (unsigned char *) nstr->orig, d = d0, d_ = NULL; ; s += ls) {
+    if (!(ls = sen_str_charlen_utf8(s, e))) {
+      break;
+    }
     if ((p = (unsigned char *)sen_nfkc_map1(s))) {
       pe = p + strlen((char *)p);
     } else {
@@ -589,8 +730,10 @@ normalize_utf8(sen_nstr *nstr)
       d = d_;
       length--;
     }
-    for (; *p && p < pe; p += lp) {
-      lp = sen_str_charlen((char *)p, sen_enc_utf8);
+    for (; ; p += lp) {
+      if (!(lp = sen_str_charlen_utf8(p, pe))) {
+        break;
+      }
       if ((*p == ' ' && removeblankp)
           || *p < 0x20  /* skip unprintable ascii */ ) {
         if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
@@ -617,6 +760,7 @@ normalize_utf8(sen_nstr *nstr)
   if (cp) { *cp = sen_str_null; }
   *d = '\0';
   nstr->length = length;
+  nstr->norm_blen = (size_t)(d - (unsigned char *)nstr->norm);
   return sen_success;
 }
 #endif /* NO_NFKC */
@@ -645,9 +789,9 @@ normalize_sjis(sen_nstr *nstr)
   };
   int_least8_t *ch;
   const unsigned char *s, *s_;
-  unsigned char *d, *d0, *d_, b;
+  unsigned char *d, *d0, *d_, b, *e;
   uint_least8_t *cp, *ctypes, ctype;
-  size_t size = strlen(nstr->orig), length = 0;
+  size_t size = nstr->orig_blen, length = 0;
   int removeblankp = nstr->flags & SEN_STR_REMOVEBLANK;
   if (!(nstr->norm = SEN_MALLOC(size * 2 + 1))) {
     return sen_memory_exhausted;
@@ -668,7 +812,8 @@ normalize_sjis(sen_nstr *nstr)
     }
   }
   cp = ctypes = nstr->ctypes;
-  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; *s; s++) {
+  e = (unsigned char *)nstr->orig + size;
+  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; s < e; s++) {
     if ((*s & 0x80)) {
       if (0xa0 <= *s && *s <= 0xdf) {
         uint16_t c = hankana[*s - 0xa0];
@@ -699,7 +844,7 @@ normalize_sjis(sen_nstr *nstr)
         }
         ctype = sen_str_katakana;
       } else {
-        if (0x40 <= *(s + 1) && *(s + 1) <= 0xfc) {
+        if ((s + 1) < e && 0x40 <= *(s + 1) && *(s + 1) <= 0xfc) {
           unsigned char c1 = *s++, c2 = *s, c3 = 0;
           if (0x81 <= c1 && c1 <= 0x87) {
             switch (c1 & 0x0f) {
@@ -843,6 +988,7 @@ normalize_sjis(sen_nstr *nstr)
   if (cp) { *cp = sen_str_null; }
   *d = '\0';
   nstr->length = length;
+  nstr->norm_blen = (size_t)(d - (unsigned char *)nstr->norm);
   return sen_success;
 }
 
@@ -850,10 +996,10 @@ inline static sen_rc
 normalize_none(sen_nstr *nstr)
 {
   int_least8_t *ch;
-  const unsigned char *s, *s_;
+  const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_;
   uint_least8_t *cp, *ctypes, ctype;
-  size_t size = strlen(nstr->orig), length = 0;
+  size_t size = nstr->orig_blen, length = 0;
   int removeblankp = nstr->flags & SEN_STR_REMOVEBLANK;
   if (!(nstr->norm = SEN_MALLOC(size + 1))) {
     return sen_memory_exhausted;
@@ -874,7 +1020,8 @@ normalize_none(sen_nstr *nstr)
     }
   }
   cp = ctypes = nstr->ctypes;
-  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; *s; s++) {
+  e = (unsigned char *)nstr->orig + size;
+  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; s < e; s++) {
     unsigned char c = *s;
     switch (c >> 4) {
     case 0 :
@@ -933,6 +1080,7 @@ normalize_none(sen_nstr *nstr)
   if (cp) { *cp = sen_str_null; }
   *d = '\0';
   nstr->length = length;
+  nstr->norm_blen = (size_t)(d - (unsigned char *)nstr->norm);
   return sen_success;
 }
 
@@ -941,7 +1089,7 @@ inline static sen_rc
 normalize_latin1(sen_nstr *nstr)
 {
   int_least8_t *ch;
-  const unsigned char *s, *s_;
+  const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_;
   uint_least8_t *cp, *ctypes, ctype;
   size_t size = strlen(nstr->orig), length = 0;
@@ -965,7 +1113,8 @@ normalize_latin1(sen_nstr *nstr)
     }
   }
   cp = ctypes = nstr->ctypes;
-  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; *s; s++) {
+  e = (unsigned char *)nstr->orig + size;
+  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; s < e; s++) {
     unsigned char c = *s;
     switch (c >> 4) {
     case 0 :
@@ -1058,6 +1207,7 @@ normalize_latin1(sen_nstr *nstr)
   if (cp) { *cp = sen_str_null; }
   *d = '\0';
   nstr->length = length;
+  nstr->norm_blen = (size_t)(d - (unsigned char *)nstr->norm);
   return sen_success;
 }
 
@@ -1065,7 +1215,7 @@ inline static sen_rc
 normalize_koi8r(sen_nstr *nstr)
 {
   int_least8_t *ch;
-  const unsigned char *s, *s_;
+  const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_;
   uint_least8_t *cp, *ctypes, ctype;
   size_t size = strlen(nstr->orig), length = 0;
@@ -1089,7 +1239,8 @@ normalize_koi8r(sen_nstr *nstr)
     }
   }
   cp = ctypes = nstr->ctypes;
-  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; *s; s++) {
+  e = (unsigned char *)nstr->orig + size;
+  for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; s < e; s++) {
     unsigned char c = *s;
     switch (c >> 4) {
     case 0 :
@@ -1171,18 +1322,21 @@ normalize_koi8r(sen_nstr *nstr)
   if (cp) { *cp = sen_str_null; }
   *d = '\0';
   nstr->length = length;
+  nstr->norm_blen = (size_t)(d - (unsigned char *)nstr->norm);
   return sen_success;
 }
 
 sen_nstr *
-sen_nstr_open(const char *str, sen_encoding encoding, int flags)
+sen_nstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags)
 {
   sen_rc rc;
   sen_nstr *nstr;
   if (!str || !*str) { return NULL; }
   if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) { return NULL; }
   nstr->orig = str;
+  nstr->orig_blen = str_len;
   nstr->norm = NULL;
+  nstr->norm_blen = 0;
   nstr->checks = NULL;
   nstr->ctypes = NULL;
   nstr->flags = flags;
@@ -1217,6 +1371,89 @@ sen_nstr_open(const char *str, sen_encoding encoding, int flags)
   return nstr;
 }
 
+sen_nstr *
+sen_fakenstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags)
+{
+  /* TODO: support SEN_STR_REMOVEBLANK flag and ctypes */
+  sen_nstr *nstr;
+
+  if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) {
+    return NULL;
+  }
+  if (!(nstr->norm = SEN_MALLOC(str_len + 1))) {
+    SEN_LOG(sen_log_alert, "memory allocation for keyword on sen_snip_add_cond failed !");
+    SEN_FREE(nstr);
+    return NULL;
+  }
+  nstr->orig = str;
+  nstr->orig_blen = str_len;
+  memcpy(nstr->norm, str, str_len);
+  nstr->norm[str_len] = '\0';
+  nstr->norm_blen = str_len;
+  nstr->ctypes = NULL;
+
+  if (flags & SEN_STR_WITH_CHECKS) {
+    int_least8_t f = 0;
+    unsigned char c;
+    size_t i;
+    if (!(nstr->checks = (int_least8_t *) SEN_MALLOC(sizeof(int_least8_t) * str_len))) {
+      SEN_FREE(nstr->norm);
+      SEN_FREE(nstr);
+      return NULL;
+    }
+    switch (encoding) {
+    case sen_enc_euc_jp:
+      for (i = 0; i < str_len; i++) {
+        if (!f) {
+          c = (unsigned char) str[i];
+          f = ((c >= 0xa1U && c <= 0xfeU) || c == 0x8eU ? 2 : (c == 0x8fU ? 3 : 1)
+            );
+          nstr->checks[i] = f;
+        } else {
+          nstr->checks[i] = 0;
+        }
+        f--;
+      }
+      break;
+    case sen_enc_sjis:
+      for (i = 0; i < str_len; i++) {
+        if (!f) {
+          c = (unsigned char) str[i];
+          f = (c >= 0x81U && ((c <= 0x9fU) || (c >= 0xe0U && c <= 0xfcU)) ? 2 : 1);
+          nstr->checks[i] = f;
+        } else {
+          nstr->checks[i] = 0;
+        }
+        f--;
+      }
+      break;
+    case sen_enc_utf8:
+      for (i = 0; i < str_len; i++) {
+        if (!f) {
+          c = (unsigned char) str[i];
+          f = (c & 0x80U ? (c & 0x20U ? (c & 0x10U ? 4 : 3)
+                           : 2)
+               : 1);
+          nstr->checks[i] = f;
+        } else {
+          nstr->checks[i] = 0;
+        }
+        f--;
+      }
+      break;
+    default:
+      for (i = 0; i < str_len; i++) {
+        nstr->checks[i] = 1;
+      }
+      break;
+    }
+  }
+  else {
+    nstr->checks = NULL;
+  }
+  return nstr;
+}
+
 sen_rc
 sen_nstr_close(sen_nstr *nstr)
 {
@@ -1238,57 +1475,17 @@ const char * sen_enc_string[] = {
   "koi8r"
 };
 
-size_t
-sen_str_charlen(const char *str, sen_encoding encoding)
-{
-  size_t size;
-  const unsigned char *p = (const unsigned char *) str;
-  if (!*p) { return 0; }
-  switch (encoding) {
-  case sen_enc_euc_jp :
-    size = ((*p & 0x80) && (*(p + 1))) ? 2 : 1;
-    break;
-  case sen_enc_utf8 :
-    if (*p & 0x80) {
-      int b, w; 
-      for (b = 0x40, w = 0; b && (*p & b); b >>= 1, w++);
-      if (!w) { SEN_LOG(sen_log_warning, "invalid utf8 string"); }
-      for (size = 1; w--; size++) {
-        if (!*++p || (*p & 0xc0) != 0x80) {
-          SEN_LOG(sen_log_warning, "invalid utf8 string");
-          break;
-        }
-      }
-    } else {
-      size = 1;
-    }
-    break;
-  case sen_enc_sjis :
-    if (*p & 0x80) {
-      if ((0xa1 <= *p && *p <= 0xdf) || !(*(p + 1))) {
-        size = 1;
-      } else {
-        size = 2;
-      }
-    } else {
-      size = 1;
-    }
-    break;
-  default :
-    size = 1;
-    break;
-  }
-  return size;
-}
-
 size_t
 sen_str_len(const char *str, sen_encoding encoding, const char **last)
 {
-  size_t len;
+  size_t len, tlen;
   const char *p = NULL;
-  for (len = 0; *str; len++) {
+  for (len = 0; ; len++) {
     p = str;
-    str += sen_str_charlen(str, encoding);
+    if (!(tlen = sen_str_charlen(str, encoding))) {
+      break;
+    }
+    str += tlen;
   }
   if (last) { *last = p; }
   return len;
@@ -1314,7 +1511,7 @@ sen_isspace(const char *str, sen_encoding encoding)
     if (encoding == sen_enc_euc_jp && s[1] == 0xA1) { return 1; }
     break;
   case 0xE3 :
-    if (encoding == sen_enc_utf8 && s[1] == 0x80 && s[2] == 0x80) { return 1; } 
+    if (encoding == sen_enc_utf8 && s[1] == 0x80 && s[2] == 0x80) { return 1; }
     break;
   default :
     break;
