@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -52,7 +52,9 @@ typedef struct _sen_str_splitter sen_str_splitter;
 
 typedef struct {
   const char *orig;
+  size_t orig_blen;
   char *norm;
+  size_t norm_blen;
   uint_least8_t *ctypes;
   int_least8_t *checks;
   size_t length;
@@ -68,11 +70,12 @@ int sen_str_split(char *str, sen_substr *res, int max, char **rest);
 
 int sen_str_get_prefix_order(const char *str);
 
-
-sen_nstr * sen_nstr_open(const char *str, sen_encoding encoding, int flags);
+sen_nstr *sen_nstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags);
+sen_nstr *sen_fakenstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags);
 sen_rc sen_nstr_close(sen_nstr *nstr);
 
 size_t sen_str_charlen(const char *str, sen_encoding encoding);
+size_t sen_str_charlen_nonnull(const char *str, const char *end, sen_encoding encoding);
 size_t sen_str_len(const char *str, sen_encoding encoding, const char **last);
 sen_rc sen_str_fin(void);
 
@@ -84,7 +87,7 @@ void *sen_malloc(size_t size, const char* file, int line);
 void *sen_calloc(size_t size, const char* file, int line);
 void sen_free(void *ptr, const char* file, int line);
 void *sen_realloc(void *ptr, size_t size, const char* file, int line);
-char * sen_strdup(const char *s, const char* file, int line);
+char *sen_strdup(const char *s, const char* file, int line);
 void sen_assert(int cond, const char* file, int line, const char* func);
 int sen_isspace(const char *s, sen_encoding encoding);
 
@@ -96,7 +99,7 @@ int sen_isspace(const char *s, sen_encoding encoding);
 #ifdef DEBUG
 #define SEN_ASSERT(s) sen_assert((s),__FILE__,__LINE__,__FUNCTION__)
 #else
-#define SEN_ASSERT(s) 
+#define SEN_ASSERT(s)
 #endif
 
 #ifdef __cplusplus
