@@ -16,13 +16,30 @@
 */
 #include "senna_in.h"
 #include <fcntl.h>
-#include <stdio.h> /* for debug */
 #include <string.h>
 #include <sys/stat.h>
 #include "sym.h"
 
+/* v08 functions */
+
+sen_sym *sen_sym_create08(const char *path, uint32_t key_size,
+                          uint32_t flags, sen_encoding encoding);
+sen_sym *sen_sym_open08(const char *path);
+sen_id sen_sym_get08(sen_sym *sym, const void *key);
+sen_id sen_sym_at08(sen_sym *sym, const void *key);
+sen_set *sen_sym_prefix_search08(sen_sym *sym, const void *key);
+sen_set *sen_sym_suffix_search08(sen_sym *sym, const void *key);
+sen_id sen_sym_common_prefix_search08(sen_sym *sym, const void *key);
+sen_rc sen_sym_del08(sen_sym *sym, const void *key);
+const char *_sen_sym_key08(sen_sym *sym, sen_id id);
+int sen_sym_key08(sen_sym *sym, sen_id id, void *keybuf, int bufsize);
+int sen_sym_pocket_get08(sen_sym *sym, sen_id id);
+sen_rc sen_sym_pocket_set08(sen_sym *sym, sen_id id, unsigned int value);
+int sen_sym_del_with_sis08(sen_sym *sym, sen_id id,
+                           int(*func)(sen_id, void *), void *func_arg);
+
 #define SEN_SYM_HEADER_SIZE 0x10000
-#define SEN_IDSTR "SENNA:SYM:00.00"
+#define SEN_IDSTR "SENNA:SYM:01.00"
 #define SEN_SYM_DELETED (SEN_SYM_MAX_ID + 1)
 
 #define SEN_SYM_SEGMENT_SIZE 0x400000
@@ -32,6 +49,27 @@
 #define KEY_MASK_IN_A_SEGMENT 0x3fffff
 #define PAT_MASK_IN_A_SEGMENT 0x3ffff
 #define SIS_MASK_IN_A_SEGMENT 0x7ffff
+#define SEG_NOT_ASSIGNED 0xffff
+
+struct sen_sym_header {
+  char idstr[16];
+  uint32_t flags;
+  sen_encoding encoding;
+  uint32_t key_size;
+  uint32_t nrecords;
+  uint32_t curr_rec;
+  int32_t curr_key;
+  int32_t curr_del;
+  int32_t curr_del2;
+  int32_t curr_del3;
+  uint32_t lock;
+  uint32_t reserved[18];
+  uint16_t keyarray[SEN_SYM_MAX_SEGMENT];
+  uint16_t patarray[SEN_SYM_MAX_SEGMENT];
+  uint16_t sisarray[SEN_SYM_MAX_SEGMENT];
+  syn_sym_delinfo delinfos[SEN_SYM_NDELINFOS];
+  sen_id garbages[(SEN_SYM_MAX_KEY_LENGTH + 1) / 8];
+};
 
 typedef struct {
   sen_id r;
@@ -67,69 +105,53 @@ enum {
 /* bit operation */
 
 inline static uint32_t
-nth_bit(const uint8_t *key, uint32_t n)
+nth_bit(const byte *key, uint32_t n)
 {
   return key[n >> 3] & (0x80 >> (n & 7));
 }
 
 /* segment operation */
 
-inline static sen_rc
-load_all_segments(sen_sym *sym)
-{
-  sen_rc rc = sen_success;
-  int seg, nkey = 0, npat = 0, nsis = 0, empty = 0;
-  for (seg = 0; seg < SEN_SYM_MAX_SEGMENT; seg++) {
-    switch (sym->header->segments[seg]) {
-    case segment_empty :
-      empty++;
-      break;
-    case segment_key :
-      sym->keyarray[nkey++].segno = seg;
-      break;
-    case segment_pat :
-      sym->patarray[npat++].segno = seg;
-      break;
-    case segment_sis :
-      sym->sisarray[nsis++].segno = seg;
-      break;
-    default :
-      rc = sen_invalid_format;
-      break;
-    }
-  }
-  return rc;
-}
-
 inline static sen_rc
 segment_new(sen_sym *sym, int segtype)
 {
-  int seg = 0, type, n = 0;
-  while ((type = sym->header->segments[seg])) {
-    if (type == segtype) { n++; }
-    if (++seg >= SEN_SYM_MAX_SEGMENT) { return sen_memory_exhausted; }
+  int i;
+  uint16_t seg = 0, k = 0, p = 0, s = 0;
+  char used[SEN_SYM_MAX_SEGMENT];
+  memset(used, 0, SEN_SYM_MAX_SEGMENT);
+  for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
+    if ((seg = sym->header->keyarray[i]) != SEG_NOT_ASSIGNED) {
+      k = i + 1; used[seg] = 1;
+    }
+    if ((seg = sym->header->patarray[i]) != SEG_NOT_ASSIGNED) {
+      p = i + 1; used[seg] = 1;
+    }
+    if ((seg = sym->header->sisarray[i]) != SEG_NOT_ASSIGNED) {
+      s = i + 1; used[seg] = 1;
+    }
+  }
+  for (seg = 0; used[seg];) {
+    if (++seg == SEN_SYM_MAX_SEGMENT) { return sen_memory_exhausted; }
   }
-  sym->header->segments[seg] = segtype;
   switch (segtype) {
   case segment_key :
-    sym->keyarray[n].segno = seg;
-    sym->header->curr_key = n * SEN_SYM_SEGMENT_SIZE;
+    sym->header->keyarray[k] = seg;
+    sym->header->curr_key = k * SEN_SYM_SEGMENT_SIZE;
     break;
   case segment_pat :
-    sym->patarray[n].segno = seg;
+    sym->header->patarray[p] = seg;
     break;
   case segment_sis :
-    sym->sisarray[n].segno = seg;
+    sym->header->sisarray[s] = seg;
     break;
   }
   return sen_success;
 }
 
-#define SEG_LOAD(io,si)\
+#define SEG_LOAD(io,segno,addr)\
 {\
-  int32_t segno = si->segno;\
-  if (0 <= segno && segno < SEN_SYM_MAX_SEGMENT) {\
-    SEN_IO_SEG_REF(io, segno, si->addr);\
+  if (segno < SEN_SYM_MAX_SEGMENT) {\
+    SEN_IO_SEG_REF(io, segno, addr);\
     SEN_IO_SEG_UNREF(io, segno);\
   }\
 }
@@ -139,34 +161,36 @@ segment_new(sen_sym *sym, int segtype)
 inline static pat_node *
 pat_at(sen_sym *sym, sen_id id)
 {
-  sen_sym_seginfo *si;
+  uint16_t lseg, pseg;
+  void **p;
   if (id > SEN_SYM_MAX_ID) { return NULL; }
-  si = &sym->patarray[id >> W_OF_PAT_IN_A_SEGMENT];
-  if (!si->addr) {
-    if (si->segno == -1) { load_all_segments(sym); }
-    SEG_LOAD(sym->io, si);
-    if (!si->addr) { return NULL; }
+  lseg = id >> W_OF_PAT_IN_A_SEGMENT;
+  p = &sym->pataddrs[lseg];
+  if (!*p) {
+    if ((pseg = sym->header->patarray[lseg]) == SEG_NOT_ASSIGNED) { return NULL; }
+    SEG_LOAD(sym->io, pseg, *p);
+    if (!*p) { return NULL; }
   }
-  return
-    (pat_node *)(((byte *)si->addr) + ((id & PAT_MASK_IN_A_SEGMENT) * sizeof(pat_node)));
+  return (pat_node *)(((byte *)*p) + ((id & PAT_MASK_IN_A_SEGMENT) * sizeof(pat_node)));
 }
 
+
 inline static pat_node *
 pat_get(sen_sym *sym, sen_id id)
 {
-  sen_sym_seginfo *si;
+  uint16_t lseg, pseg;
+  void **p;
   if (id > SEN_SYM_MAX_ID) { return NULL; }
-  si = &sym->patarray[id >> W_OF_PAT_IN_A_SEGMENT];
-  if (!si->addr) {
-    if (si->segno == -1) { load_all_segments(sym); }
-    while (si->segno == -1) {
-      if (segment_new(sym, segment_pat)) { return NULL; }
+  lseg = id >> W_OF_PAT_IN_A_SEGMENT;
+  p = &sym->pataddrs[lseg];
+  if (!*p) {
+    while ((pseg = sym->header->patarray[lseg]) == SEG_NOT_ASSIGNED) {
+      segment_new(sym, segment_pat);
     }
-    SEG_LOAD(sym->io, si);
-    if (!si->addr) { return NULL; }
+    SEG_LOAD(sym->io, pseg, *p);
+    if (!*p) { return NULL; }
   }
-  return
-    (pat_node *)(((byte *)si->addr) + ((id & PAT_MASK_IN_A_SEGMENT) * sizeof(pat_node)));
+  return (pat_node *)(((byte *)*p) + ((id & PAT_MASK_IN_A_SEGMENT) * sizeof(pat_node)));
 }
 
 inline static pat_node *
@@ -188,34 +212,35 @@ pat_node_new(sen_sym *sym, sen_id *id)
 inline static sis_node *
 sis_at(sen_sym *sym, sen_id id)
 {
-  sen_sym_seginfo *si;
+  uint16_t lseg, pseg;
+  void **p;
   if (id > SEN_SYM_MAX_ID) { return NULL; }
-  si = &sym->sisarray[id >> W_OF_SIS_IN_A_SEGMENT];
-  if (!si->addr) {
-    if (si->segno == -1) { load_all_segments(sym); }
-    SEG_LOAD(sym->io, si);
-    if (!si->addr) { return NULL; }
+  lseg = id >> W_OF_SIS_IN_A_SEGMENT;
+  p = &sym->sisaddrs[lseg];
+  if (!*p) {
+    if ((pseg = sym->header->sisarray[lseg]) == SEG_NOT_ASSIGNED) { return NULL; }
+    SEG_LOAD(sym->io, pseg, *p);
+    if (!*p) { return NULL; }
   }
-  return
-    (sis_node *)(((byte *)si->addr) + ((id & SIS_MASK_IN_A_SEGMENT) * sizeof(sis_node)));
+  return (sis_node *)(((byte *)*p) + ((id & SIS_MASK_IN_A_SEGMENT) * sizeof(sis_node)));
 }
 
 inline static sis_node *
 sis_get(sen_sym *sym, sen_id id)
 {
-  sen_sym_seginfo *si;
-  if (id == SEN_SYM_NIL || SEN_SYM_MAX_ID < id) { return NULL; }
-  si = &sym->sisarray[id >> W_OF_SIS_IN_A_SEGMENT];
-  if (!si->addr) {
-    if (si->segno == -1) { load_all_segments(sym); }
-    while (si->segno == -1) {
-      if (segment_new(sym, segment_sis)) { return NULL; }
+  uint16_t lseg, pseg;
+  void **p;
+  if (id > SEN_SYM_MAX_ID) { return NULL; }
+  lseg = id >> W_OF_SIS_IN_A_SEGMENT;
+  p = &sym->sisaddrs[lseg];
+  if (!*p) {
+    while ((pseg = sym->header->sisarray[lseg]) == SEG_NOT_ASSIGNED) {
+      segment_new(sym, segment_sis);
     }
-    SEG_LOAD(sym->io, si);
-    if (!si->addr) { return NULL; }
+    SEG_LOAD(sym->io, pseg, *p);
+    if (!*p) { return NULL; }
   }
-  return
-    (sis_node *)(((byte *)si->addr) + ((id & SIS_MASK_IN_A_SEGMENT) * sizeof(sis_node)));
+  return (sis_node *)(((byte *)*p) + ((id & SIS_MASK_IN_A_SEGMENT) * sizeof(sis_node)));
 }
 
 #define MAX_LEVEL 16
@@ -241,15 +266,19 @@ sis_collect(sen_sym *sym, sen_set *h, sen_id id, uint32_t level)
 
 #define KEY_AT(sym,pos,ptr) \
 {\
-  sen_sym_seginfo *si = &sym->keyarray[pos >> W_OF_KEY_IN_A_SEGMENT];\
-  if (si->addr) {\
-    ptr = (char *)(((byte *)si->addr) + (pos & KEY_MASK_IN_A_SEGMENT));\
+  uint16_t lseg, pseg;\
+  void **p;\
+  lseg = pos >> W_OF_KEY_IN_A_SEGMENT;\
+  p = &sym->keyaddrs[lseg];\
+  if (*p) {\
+    ptr = ((byte *)*p) + (pos & KEY_MASK_IN_A_SEGMENT);\
   } else {\
-    if (si->segno == -1) { load_all_segments(sym); }\
-    SEG_LOAD(sym->io, si);\
-    ptr = si->addr\
-      ? (char *)(((byte *)si->addr) + (pos & KEY_MASK_IN_A_SEGMENT))\
-      : NULL;\
+    if ((pseg = sym->header->keyarray[lseg]) == SEG_NOT_ASSIGNED) {\
+      ptr = NULL;\
+    } else {\
+      SEG_LOAD(sym->io, pseg, *p);\
+      ptr = *p ? ((byte *)*p) + (pos & KEY_MASK_IN_A_SEGMENT) : NULL;\
+    }\
   }\
 }
 
@@ -265,7 +294,7 @@ key_put(sen_sym *sym, const uint8_t *key, int len)
     res = sym->header->curr_key;
   }
   {
-    char *dest;
+    uint8_t *dest;
     KEY_AT(sym, res, dest);
     if (!dest) { return 0; }
     memcpy(dest, key, len);
@@ -274,13 +303,13 @@ key_put(sen_sym *sym, const uint8_t *key, int len)
   return res;
 }
 
-inline static char *
+inline static uint8_t *
 pat_node_get_key(sen_sym *sym, pat_node *n)
 {
   if (PAT_IMD(n)) {
-    return (char *)&n->key;
+    return (uint8_t *) &n->key;
   } else {
-    char *res;
+    uint8_t *res;
     KEY_AT(sym, n->key, res);
     return res;
   }
@@ -305,13 +334,15 @@ pat_node_set_key(sen_sym *sym, pat_node *n, const uint8_t *key, unsigned int len
 #define DL_EMPTY  0x00000000
 #define DL_PHASE1 0x00000001
 #define DL_PHASE2 0x00000002
+#define DL_STMASK 0x00000003
 #define DL_SHARED 0x80000000
 #define DL_LD(x) (((x)->bitfield >> 2) & SEN_SYM_MAX_ID)
-#define DL_ST(x) ((x)->bitfield & 3)
+#define DL_ST(x) ((x)->bitfield & DL_STMASK)
 #define DL_SH(x) ((x)->bitfield & DL_SHARED)
-#define SET_DL_LD(x,v) ((x)->bitfield = DL_ST(x) | ((v) << 2))
-#define SET_DL_ST(x,v) ((x)->bitfield = ((x)->bitfield & 0xfffffffc) | ((v) & 3))
-#define SET_DL_SH(x,v) ((x)->bitfield = ((x)->bitfield & ~DL_SHARED) | (v))
+#define SET_DL_LD(x,v) ((x)->bitfield = ((x)->bitfield & (DL_STMASK|DL_SHARED)) | \
+                                        (((v) & SEN_SYM_MAX_ID) << 2))
+#define SET_DL_ST(x,v) ((x)->bitfield = ((x)->bitfield & ~DL_STMASK) | ((v) & DL_STMASK))
+#define SET_DL_SH(x,v) ((x)->bitfield = ((x)->bitfield & ~DL_SHARED) | ((v) & DL_SHARED))
 
 inline static syn_sym_delinfo *
 delinfo_search(sen_sym *sym, sen_id id)
@@ -346,9 +377,9 @@ delinfo_turn_2(sen_sym *sym, syn_sym_delinfo *di)
     pat_node *rn;
     int c;
     size_t len = sym->key_size * 8;
-    const char *key = pat_node_get_key(sym, dn);
+    const uint8_t *key = pat_node_get_key(sym, dn);
     if (!key) { return sen_invalid_argument; }
-    if (!len) { len = (strlen(key) + 1) * 8; }
+    if (!len) { len = (strlen((char *)key) + 1) * 8; }
     rn = pat_at(sym, 0);
     c = -1;
     p0 = &rn->r;
@@ -411,9 +442,9 @@ delinfo_turn_3(sen_sym *sym, syn_sym_delinfo *di)
       if (PAT_IMD(dn)) {
         size = 0;
       } else {
-        const char *key = pat_node_get_key(sym, dn);
+        const uint8_t *key = pat_node_get_key(sym, dn);
         if (!key) { return sen_invalid_argument; }
-        size = strlen(key) + 1;
+        size = strlen((char *)key) + 1;
       }
     }
   }
@@ -458,6 +489,9 @@ sen_sym_create(const char *path, uint32_t key_size,
   sen_sym *sym = NULL;
   pat_node *node0;
   struct sen_sym_header *header;
+  if (flags & 0x70000) {
+    return sen_sym_create08(path, key_size, flags, encoding);
+  }
   io = sen_io_create(path, SEN_SYM_HEADER_SIZE, SEN_SYM_SEGMENT_SIZE,
                      SEN_SYM_MAX_SEGMENT, sen_io_auto, SEN_SYM_MAX_SEGMENT);
   if (!io) { return NULL; }
@@ -474,24 +508,25 @@ sen_sym_create(const char *path, uint32_t key_size,
   header->curr_del3 = 0;
   header->lock = 0;
   for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
-    header->segments[i] = segment_empty;
+    header->keyarray[i] = SEG_NOT_ASSIGNED;
+    header->patarray[i] = SEG_NOT_ASSIGNED;
+    header->sisarray[i] = SEG_NOT_ASSIGNED;
   }
   if (!(sym = SEN_MALLOC(sizeof(sen_sym)))) {
     sen_io_close(io);
     return NULL;
   }
+  sym->v08p = 0;
   sym->io = io;
   sym->header = header;
   sym->key_size = key_size;
   sym->encoding = encoding;
   sym->flags = flags;
+  sym->lock = &header->lock;
   for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
-    sym->keyarray[i].segno = -1;
-    sym->keyarray[i].addr = NULL;
-    sym->patarray[i].segno = -1;
-    sym->patarray[i].addr = NULL;
-    sym->sisarray[i].segno = -1;
-    sym->sisarray[i].addr = NULL;
+    sym->keyaddrs[i] = NULL;
+    sym->pataddrs[i] = NULL;
+    sym->sisaddrs[i] = NULL;
   }
   node0 = pat_get(sym, 0);
   node0->r = 0;
@@ -510,32 +545,27 @@ sen_sym_open(const char *path)
   io = sen_io_open(path, sen_io_auto, 8192);
   if (!io) { return NULL; }
   header = sen_io_header(io);
-  /*
   if (memcmp(header->idstr, SEN_IDSTR, 16)) {
-    puts("id not match");
+    SEN_LOG(sen_log_notice, "sym_idstr (%s)", header->idstr);
     sen_io_close(io);
-    return NULL;
+    return sen_sym_open08(path);
   }
-  */
   if (!(sym = SEN_MALLOC(sizeof(sen_sym)))) {
-    puts("malloc failed");
     sen_io_close(io);
     return NULL;
   }
+  sym->v08p = 0;
   sym->io = io;
   sym->header = header;
   sym->key_size = header->key_size;
   sym->encoding = header->encoding;
   sym->flags = header->flags;
+  sym->lock = &header->lock;
   for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
-    sym->keyarray[i].segno = -1;
-    sym->keyarray[i].addr = NULL;
-    sym->patarray[i].segno = -1;
-    sym->patarray[i].addr = NULL;
-    sym->sisarray[i].segno = -1;
-    sym->sisarray[i].addr = NULL;
+    sym->keyaddrs[i] = NULL;
+    sym->pataddrs[i] = NULL;
+    sym->sisaddrs[i] = NULL;
   }
-  load_all_segments(sym);
   return sym;
 }
 
@@ -565,14 +595,13 @@ _sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new, uint32_t *lkey)
   int c = -1, c0 = -1, c1 = -1, len;
   size_t size = sym->key_size;
   *new = 0;
-  if (!key) { return SEN_SYM_NIL; }
   if (!size) { size = strlen((char *)key) + 1; }
   len = (int)size * 8;
   if (len > SEN_SYM_MAX_KEY_LENGTH) { return SEN_SYM_NIL; }
   rn0 = pat_at(sym, 0); p0 = &rn0->r;
   if (*p0) {
     char xor, mask;
-    const char *s, *d;
+    const uint8_t *s, *d;
     while ((r0 = *p0)) {
       if (!(rn0 = pat_at(sym, r0))) { return 0; }
       if (c0 < rn0->check && rn0->check < len) {
@@ -585,13 +614,13 @@ _sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new, uint32_t *lkey)
         break;
       }
     }
-    for (c = 0, s = pat_node_get_key(sym, rn0), d = (char *)key; *s == *d; c += 8, s++, d++);
+    for (c = 0, s = pat_node_get_key(sym, rn0), d = key; *s == *d; c += 8, s++, d++);
     for (xor = *s ^ *d, mask = 0x80; !(xor & mask); mask >>= 1, c++);
-    // for test
-    if (r0 && r0 != sen_sym_at(sym, pat_node_get_key(sym, rn0))) {
+    /* for test
+e    if (r0 && r0 != sen_sym_at(sym, pat_node_get_key(sym, rn0))) {
       SEN_LOG(sen_log_debug, "deleted node is used as ld %d(%d:%s) %d", r0, PAT_DEL(rn0), pat_node_get_key(sym, rn0), sen_sym_at(sym, pat_node_get_key(sym, rn0)));
     }
-
+    */
     if (c == c0 && !*p0) {
       if (c < len - 1) { c++; }
     } else {
@@ -610,7 +639,7 @@ _sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new, uint32_t *lkey)
     }
     if (c >= len) { return 0; }
   } else {
-    c = (len - 1 > SEN_SYM_MAX_KEY_LENGTH) ? SEN_SYM_MAX_KEY_LENGTH : len - 1;
+    c = len - 1;
   }
   {
     size_t size2 = size > sizeof(uint32_t) ? size : 0;
@@ -672,6 +701,8 @@ sen_sym_get(sen_sym *sym, const void *key)
 {
   uint32_t new, lkey = 0;
   sen_id r0;
+  if (!key) { return SEN_SYM_NIL; }
+  if (sym->v08p) { return sen_sym_get08(sym, key); }
   r0 = _sen_sym_get(sym, (uint8_t *)key, &new, &lkey);
   if ((sym->flags & SEN_SYM_WITH_SIS) && (*((uint8_t *)key) & 0x80)) { // todo: refine!!
     sis_node *sl, *sr;
@@ -708,6 +739,7 @@ sen_sym_at(sen_sym *sym, const void *key)
   pat_node *rn;
   int c = -1;
   size_t len = sym->key_size * 8;
+  if (sym->v08p) { return sen_sym_at08(sym, key); }
   if (!key) { return SEN_SYM_NIL; }
   if (!len) { len = (strlen(key) + 1) * 8; }
   for (r = pat_at(sym, 0)->r; r; r = nth_bit((uint8_t *)key, c) ? rn->r : rn->l) {
@@ -754,6 +786,7 @@ sen_sym_prefix_search(sen_sym *sym, const void *key)
   size_t len;
   sen_id r;
   pat_node *rn;
+  if (sym->v08p) { return sen_sym_prefix_search08(sym, key); }
   if (!key || sym->key_size) { return NULL; } /* only for string */
   len = strlen(key) * 8;
   rn = pat_at(sym, 0);
@@ -784,6 +817,7 @@ sen_sym_suffix_search(sen_sym *sym, const void *key)
 {
   sen_id r;
   uint32_t *offset;
+  if (sym->v08p) { return sen_sym_suffix_search08(sym, key); }
   if (!key || sym->key_size) { return NULL; }
   if ((r = sen_sym_at(sym, key))) {
     sen_set *h = sen_set_open(sizeof(sen_id), sizeof(uint32_t), 0);
@@ -803,6 +837,7 @@ sen_sym_common_prefix_search(sen_sym *sym, const void *key)
   pat_node *rn, *rnl;
   sen_id r, rl, r2 = SEN_SYM_NIL;
   int c = -1, cl, len = 0, n, len2 = 8, nth_bit;
+  if (sym->v08p) { return sen_sym_common_prefix_search08(sym, key); }
   if (!key) { return SEN_SYM_NIL; }
   if (sym->key_size) { return SEN_SYM_NIL; /* only for string */ }
   // len = (strlen(key) + 1) * 8;
@@ -810,8 +845,8 @@ sen_sym_common_prefix_search(sen_sym *sym, const void *key)
     if (!(rn = pat_at(sym, r))) { break; /* corrupt? */ }
     // if (len <= rn->check) { break; }
     if (rn->check <= c) {
-      char *p = pat_node_get_key(sym, rn);
-      if (!memcmp(p, key, strlen(p))) { return r; }
+      uint8_t *p = pat_node_get_key(sym, rn);
+      if (!memcmp(p, key, strlen((char *)p))) { return r; }
       break;
     }
     c = rn->check;
@@ -825,7 +860,7 @@ sen_sym_common_prefix_search(sen_sym *sym, const void *key)
           if (!(rnl = pat_at(sym, rl))) { break; /* corrupt? */ }
           if (len2 <= rnl->check) { break; }
           if (rnl->check <= cl) {
-            char *p = pat_node_get_key(sym, rnl);
+            uint8_t *p = pat_node_get_key(sym, rnl);
             if (!p[n] && !memcmp(p, key, n)) { r2 = rl; }
             break;
           }
@@ -955,6 +990,7 @@ _sen_sym_del(sen_sym *sym, const char *key, int shared)
 sen_rc
 sen_sym_del(sen_sym *sym, const void *key)
 {
+  if (sym->v08p) { return sen_sym_del08(sym, key); }
   return _sen_sym_del(sym, key, 0);
 }
 
@@ -967,20 +1003,22 @@ sen_sym_size(sen_sym *sym)
 const char *
 _sen_sym_key(sen_sym *sym, sen_id id)
 {
+  if (sym->v08p) { return _sen_sym_key08(sym, id); }
   pat_node *node = pat_at(sym, id);
   if (!node) { return NULL; }
-  return pat_node_get_key(sym, node);
+  return (const char *) pat_node_get_key(sym, node);
 }
 
 int
 sen_sym_key(sen_sym *sym, sen_id id, void *keybuf, int bufsize)
 {
   int len;
-  char *key;
+  uint8_t *key;
   pat_node *node;
+  if (sym->v08p) { return sen_sym_key08(sym, id, keybuf, bufsize); }
   if (!(node = pat_at(sym, id))) { return 0; }
   if (!(key = pat_node_get_key(sym, node))) { return 0; }
-  if (!(len = sym->key_size)) { len = (int)(strlen(key) + 1); }
+  if (!(len = sym->key_size)) { len = (int)(strlen((char *)key) + 1); }
   if (keybuf && bufsize >= len) { memcpy(keybuf, key, len); }
   return len;
 }
@@ -988,6 +1026,7 @@ sen_sym_key(sen_sym *sym, sen_id id, void *keybuf, int bufsize)
 int
 sen_sym_pocket_get(sen_sym *sym, sen_id id)
 {
+  if (sym->v08p) { return sen_sym_pocket_get08(sym, id); }
   pat_node *node = pat_at(sym, id);
   if (!node) { return -1; }
   return PAT_PKT(node);
@@ -996,6 +1035,7 @@ sen_sym_pocket_get(sen_sym *sym, sen_id id)
 sen_rc
 sen_sym_pocket_set(sen_sym *sym, sen_id id, unsigned int value)
 {
+  if (sym->v08p) { return sen_sym_pocket_set08(sym, id, value); }
   pat_node *node = pat_at(sym, id);
   if (value > PAT_POCKET_MAX || !node) { return sen_invalid_argument; }
   SET_PAT_PKT(node, value);
@@ -1029,6 +1069,7 @@ sen_sym_del_with_sis(sen_sym *sym, sen_id id,
   int level = 0, shared;
   const char *key = NULL, *_key;
   sis_node *sp, *ss = NULL, *si = sis_at(sym, id);
+  if (sym->v08p) { return sen_sym_del_with_sis08(sym, id, func, func_arg); }
   while (id) {
     if ((si && si->children && si->children != id) || !func(id, func_arg)) { break; }
     if (!(_key = _sen_sym_key(sym, id))) { break; }
@@ -1094,183 +1135,3 @@ sen_sym_next(sen_sym *sym, sen_id id)
   }
   return SEN_SYM_NIL;
 }
-
-#ifndef NO_DEBUG
-void
-sen_sym_dump(sen_sym *sym, int to)
-{
-  sen_id i;
-  pat_node *rn;
-  if (to < 0) { to = sym->header->curr_rec; }
-  for (i = 0; i <= (unsigned int)to; i++) {
-    rn = pat_at(sym, i);
-    SEN_LOG(sen_log_debug, "%d(%p):[%d,%d](%d)\"%s\"",
-            i, rn, rn->r, rn->l, rn->check, pat_node_get_key(sym, rn));
-  }
-}
-
-sen_id
-sen_sym_inspect_key(sen_sym *sym, const uint8_t *key)
-{
-  sen_id r;
-  pat_node *rn;
-  int c = -1;
-  size_t len = sym->key_size * 8;
-  if (!key) { return SEN_SYM_NIL; }
-  if (!len) { len = (strlen((char *)key) + 1) * 8; }
-  SEN_LOG(sen_log_debug, "sen_sym_inspect_key key=%s len=%d", key, len);
-  for (r = pat_at(sym, 0)->r; r; r = nth_bit(key, c) ? rn->r : rn->l) {
-    if (!(rn = pat_at(sym, r))) { SEN_LOG(sen_log_debug, "corrupt?"); break; /* corrupt? */ }
-    SEN_LOG(sen_log_debug,
-            "%d[%d,%d] key='%s' c=%d rn->c=%d len=%d", r, rn->r, rn->l, pat_node_get_key(sym, rn), c, rn->check, len);
-    if (len <= rn->check) { break; }
-    if (rn->check <= c) {
-      SEN_LOG(sen_log_debug, "c=%d key=%s",c, pat_node_get_key(sym, rn));
-      if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
-        SEN_LOG(sen_log_debug, "found! r=%d", r);
-        return r;
-      }
-      break;
-    }
-    c = rn->check;
-  }
-  return SEN_SYM_NIL;
-}
-
-void
-sen_sym_inspect(sen_sym *sym)
-{
-  int i;
-  printf("idstr=%s\nkey_size=%d\nnrecords=%d\ncurr_rec=%d\ncurr_key=%d\ncurr_del=%d\ncurr_del2=%d\ncurr_del3=%d\n",
-         sym->header->idstr,
-         sym->header->key_size,
-         sym->header->nrecords,
-         sym->header->curr_rec,
-         sym->header->curr_key,
-         sym->header->curr_del,
-         sym->header->curr_del2,
-         sym->header->curr_del3);
-  {
-    int se = 0, sk = 0, sa = 0, ss = 0, si = 0;
-    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
-      switch (sym->header->segments[i]) {
-      case segment_empty :
-        se++;
-        break;
-      case segment_key :
-        sk++;
-        break;
-      case segment_pat :
-        sa++;
-        break;
-      case segment_sis :
-        ss++;
-        break;
-      default :
-        si++;
-      }
-    }
-    printf("seg_empty=%d, seg_string=%d, seg_array=%d, seg_sis=%d, seg_invalid=%d\n", se, sk, sa, ss, si);
-  }
-  {
-    int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
-    printf("del2=%d, del=%d\n", sym->header->curr_del2, sym->header->curr_del);
-    for (i = sym->header->curr_del2;
-         i != sym->header->curr_del;
-         i = (i + 1) & SEN_SYM_MDELINFOS) {
-      n++;
-      switch (DL_ST(&sym->header->delinfos[i])) {
-      case 0 : s0++; break;
-      case 1 : s1++; break;
-      case 2 : s2++; break;
-      case 3 : s3++; break;
-      }
-    }
-    printf("del2->del: n=%d, s0=%d, s1=%d, s2=%d, s3=%d\n", n, s0, s1, s2, s3);
-  }
-  {
-    int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
-    for (i = sym->header->curr_del3;
-         i != sym->header->curr_del2;
-         i = (i + 1) & SEN_SYM_MDELINFOS) {
-      n++;
-      switch (DL_ST(&sym->header->delinfos[i])) {
-      case 0 : s0++; break;
-      case 1 : s1++; break;
-      case 2 : s2++; break;
-      case 3 : s3++; break;
-      }
-    }
-    printf("del3->del2: n=%d, s0=%d, s1=%d, s2=%d, s3=%d\n", n, s0, s1, s2, s3);
-  }
-  {
-    int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
-    for (i = sym->header->curr_del;
-         i != sym->header->curr_del3;
-         i = (i + 1) & SEN_SYM_MDELINFOS) {
-      n++;
-      switch (DL_ST(&sym->header->delinfos[i])) {
-      case 0 : s0++; break;
-      case 1 : s1++; break;
-      case 2 : s2++; break;
-      case 3 : s3++; break;
-      }
-    }
-    printf("del->del3: n=%d, s0=%d, s1=%d, s2=%d, s3=%d\n", n, s0, s1, s2, s3);
-  }
-  {
-    unsigned int i;
-    int n = 0, nn = 0, ndel = 0, nding = 0;
-    pat_node *tn;
-    for (i = 1; i <= sym->header->curr_rec; i++) {
-      tn = pat_at(sym, i);
-      if (tn) {
-        n++;
-        if (tn->r == SEN_SYM_DELETED) { ndel++; }
-        if (PAT_DEL(tn)) { nding++; }
-      } else {
-        nn++;
-      }
-    }
-    printf("array: n=%d, nil=%d, del=%d, deleting=%d\n", n, nn, ndel, nding);
-  }
-  {
-    int i;
-    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
-      if (sym->keyarray[i].segno != -1) { printf("keyarray[%d]=%x\n", i, sym->keyarray[i].segno); }
-    }
-    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
-      if (sym->patarray[i].segno != -1) { printf("patarray[%d]=%x\n", i, sym->patarray[i].segno); }
-    }
-    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
-      if (sym->sisarray[i].segno != -1) { printf("sisarray[%d]=%x\n", i, sym->sisarray[i].segno); }
-    }
-  }
-}
-
-static void
-show_sid(sen_sym *sym, sen_id id, int level)
-{
-  sis_node *sl = sis_at(sym, id);
-  if (sl) {
-    sen_id sid = sl->children;
-    while (sid && sid != id) {
-      int i = level;
-      while(i--) { putchar(' '); }
-      puts(_sen_sym_key(sym, sid));
-      show_sid(sym, sid, level + 1);
-      if (!(sl = sis_at(sym, sid))) { break; }
-      sid = sl->sibling;
-    }
-  }
-}
-
-void
-sen_sym_inspect_term(sen_sym *sym, sen_id id)
-{
-  puts(_sen_sym_key(sym, id));
-  if (sym->flags & SEN_SYM_WITH_SIS) {
-    show_sid(sym, id, 1);
-  }  
-}
-#endif /* NO_DEBUG */
